from datetime import datetime, date
from typing import Optional, Tuple, List

from dateutil.relativedelta import relativedelta
from terracatalogueclient import Catalogue, Product
from terracatalogueclient.config import CatalogueConfig, CatalogueEnvironment

from oscars_mgmnt_clms.manipulate_manifest import should_make_manifest, LSP_SPLIT_MANIFEST, SWI_SPLIT_MANIFEST, LCC_SPLIT_MANIFEST, \
    write_results, get_manifest_path, MAX_OSCARS, MAX_YEARS



def generate_all_manifests(manifest_root_path: str,
                           index_list_path: Optional = None,
                           publication_date: Optional[Tuple[None, str]] = None) -> None:
    """
    Function that generates all manifest files and subdirectories in the manifest_root_path directory. The index_list_path
    and publication_date parameters are optional and mainly used for testing.
    """

    # Start a CGLOPS catalogue.
    catalogue = Catalogue(CatalogueConfig.from_environment(CatalogueEnvironment.CGLS))
    # Retrieve all indices from the catalogue, generate a manifest for each one.
    for product_index in get_all_product_indices(catalogue):
        # Check if a manifest file is needed. Otherwise skip the index.
        if should_make_manifest(product_index, index_list_path):
            make_and_save_manifests(manifest_root_path, catalogue, product_index, publication_date)

def get_all_product_indices(catalogue: Catalogue) -> List[str]:
    return [collection.id for collection in catalogue.get_collections()]

def make_and_save_manifests(manifest_root_path, catalogue, product_index, publication_date):
    # Edge case 1: lsp_300m_v1_yearly_geotiff has a manifest per product type x group id.
    if product_index in LSP_SPLIT_MANIFEST:
        _per_product_type_and_productGroupId(manifest_root_path, catalogue, product_index, publication_date)

    # Edge case 2: swi_12.5km_vX_static_netcdf has a manifest per group id.
    elif product_index in SWI_SPLIT_MANIFEST:
        _per_productGroupId(manifest_root_path, catalogue, product_index, publication_date)
    # Edge case 3: clms_global_lcc_100m_v3_yearly_geotiff has a manifest per year
    elif product_index in LCC_SPLIT_MANIFEST:
        _per_year(manifest_root_path, catalogue, product_index, publication_date, [2015, 2016, 2017, 2018, 2019])
    else:
        # Generate the content of the manifest files.
        products: List[str] = generate_manifest(catalogue, product_index, publication_date=publication_date)
        # Write the content of the manifest file in the right (sub)directory.
        write_results(products, get_manifest_path(manifest_root_path, product_index))

def _per_productGroupId(manifest_root_path, catalogue, product_index, publication_date):
    collection_info = [collection for collection in catalogue.get_collections() if collection.id == product_index][0]
    for productGroupId in collection_info.properties["productInformation"]["productGroupId"]:
        query = f'properties.productInformation.productGroupId:"{productGroupId}"'
        products = catalogue.get_products(collection_info.id, publicationDate=publication_date, q=query)
        hrefs = [extract_href(product) for product in products]
        write_results(hrefs, get_manifest_path(manifest_root_path, product_index, productGroupId))


def _per_product_type_and_productGroupId(manifest_root_path, catalogue, product_index, publication_date):
    collection_info = [collection for collection in catalogue.get_collections() if collection.id == product_index][0]
    for productGroupId in collection_info.properties["productInformation"]["productGroupId"]:
        for productType in collection_info.properties["productInformation"]["productType"]:
            query = (f'properties.productInformation.productGroupId:"{productGroupId}" '
                     f'AND properties.productInformation.productType:"{productType}"')
            products = catalogue.get_products(collection_info.id, publicationDate=publication_date, q=query)
            hrefs = [extract_href(product) for product in products]
            write_results(hrefs, get_manifest_path(manifest_root_path, product_index, productGroupId, productType))


def _per_year(manifest_root_path, catalogue, product_index, publication_date, years):
    collection_info = [collection for collection in catalogue.get_collections() if collection.id == product_index][0]
    for year in years:
        query = f'properties.date:"{year}-01-01T00:00:00Z"'
        products = catalogue.get_products(collection_info.id, publicationDate=publication_date, q=query)
        hrefs = [extract_href(product) for product in products]
        write_results(hrefs, get_manifest_path(manifest_root_path, product_index, productGroupId=None,
                                               productType=None, year=year))
def generate_manifest(catalogue: Catalogue, index: str, publication_date=None) -> List[str]:
    return [extract_href(product) for product in retrieve_all_products(catalogue, index, publication_date)]

def extract_href(product: Product) -> str:
    return product.properties["links"]["data"][0]["href"]

def retrieve_all_products(catalogue: Catalogue, index: str, publication_date=None) -> List[Product]:
    total_amount_products = catalogue.get_product_count(index, publicationDate=publication_date)

    # Check if the amount of products doesn't exceed the maximum for oscars. If it does, split the retrieval in years.
    if total_amount_products < MAX_OSCARS:
        products = [product for product in catalogue.get_products(index, publicationDate=publication_date)]
    else:
        products = [product for product in retrieve_all_products_per_year(catalogue, index, publication_date)]

    # Check if the amount of products is correct, otherwise raise an error.
    check_product_amount(total_amount_products, products, index=index)
    return products

def retrieve_all_products_per_year(catalogue: Catalogue, index: str, publication_date=None) -> List[Product]:
    products = []
    for years_in_past in range(MAX_YEARS):
        start_of_year = date.today() - relativedelta(years=years_in_past + 1)
        end_of_year = date.today() - relativedelta(years=years_in_past)
        new_products = retrieve_for_year(catalogue, index, start_of_year, end_of_year,
                                         publication_date=publication_date)
        products.extend(new_products)
    return products

def retrieve_for_year(catalogue, index, start_of_year, end_of_year, publication_date=None):
    expected_amount_products = catalogue.get_product_count(index, publicationDate=publication_date,
                                                           q=get_query(start_of_year, end_of_year))
    # Check if the amount of products doesn't exceed the maximum for oscars. If it does, split the retrieval in months.
    if expected_amount_products < MAX_OSCARS:
        new_products = [product for product in catalogue.get_products(index, publicationDate=publication_date,
                                                                      q=get_query(start_of_year, end_of_year))]
    else:
        new_products = get_products_per_month(catalogue, index, start_of_year, end_of_year,
                                              publication_date=publication_date)

    # Check if the amount of products is correct, otherwise raise an error.
    check_product_amount(expected_amount_products, new_products, index=index)
    return new_products

def get_query(begin_date, end_date):
    time_field = "properties.acquisitionInformation.acquisitionParameters.beginningDateTime"
    begin_date_zulu = f"{begin_date}T00:00:00.000Z"
    actual_end_date = end_date if end_date == date.today() else end_date - relativedelta(days=1)
    end_date_zulu = f"{actual_end_date}T23:59:59.999Z"
    return f"{time_field}:[{begin_date_zulu} TO {end_date_zulu}]"


def get_products_per_month(catalogue: Catalogue, index: str, start_date: datetime.date, end_date: datetime.date,
                           publication_date=None):
    all_products = []
    begin_current_date = start_date
    for _ in range(12):
        end_current_month = begin_current_date + relativedelta(months=1)
        new_products = [product for product in catalogue.get_products(index, publicationDate=publication_date,
                                                                      q=get_query(begin_current_date,
                                                                                  min(end_current_month, end_date)))]
        all_products = all_products + new_products
        begin_current_date = end_current_month

        if end_current_month > end_date:
            # No products after end_date should be retrieved.
            break

    return all_products

def check_product_amount(total_amount_products, products, index=None) -> None:
    if total_amount_products == len(products):
        return
    else:
        if total_amount_products > len(products):
            raise Exception(
                f"Missing products. Expected {total_amount_products}, got {len(products)} for index {index}.")
        else:
            raise Exception(
                f"too many products. Expected {total_amount_products}, got {len(products)} for index {index}.")
