import json
import re
from pathlib import Path
from typing import List, Optional, Union, Tuple
import os
from terracatalogueclient import Product, Catalogue
from terracatalogueclient.config import CatalogueConfig, CatalogueEnvironment

MAX_OSCARS = 10e3
MAX_YEARS = 35

CHANGING_HREFS = ["clms_global_wl_lakes_v2_daily_geojson", "clms_global_wl_rivers_v2_daily_geojson"]
SWI_SPLIT_MANIFEST = ["clms_global_swi_12.5km_v1_static_netcdf", "clms_global_swi_12.5km_v3_static_netcdf"]
LSP_SPLIT_MANIFEST = ["clms_global_lsp_300m_v1_yearly_geotiff"]
LCC_SPLIT_MANIFEST = ["clms_global_lcc_100m_v3_yearly_geotiff"]

def should_make_manifest(collection_id: str, index_list_path: Optional[str]) -> bool:
    # If an index_list_path is provided, only generate a manifest if the collection_id is in the list.
    if index_list_path is not None:
        with open(index_list_path, 'r') as f:
            indices = f.read().splitlines()
            return collection_id in indices
    else:
        return ("netcdf" in collection_id or
                "lsp" in collection_id or
                "geojson" in collection_id or
                "clms_global_lcc" in collection_id)

def get_manifest_path(manifest_root_path, product_index, productGroupId=None, productType=None, year=None):
    subdirectory_name = product_index.replace("clms_global_", "")
    return Path(
        manifest_root_path) / f"{subdirectory_name}/{get_manifest_name(product_index, productGroupId, productType, year)}"


def get_manifest_name(product_index: str, productGroupId=None, productType=None, year=None) -> str:
    manifest_name = f"manifest_{product_index}_latest.txt"
    is_edge_case = productGroupId or productType or year

    if not is_edge_case:
        return manifest_name

    # Edge case 1: lsp collection is split over multiple manifests.
    lsp_substring = "clms_global_lsp"
    if product_index in LSP_SPLIT_MANIFEST:
        split_parts = manifest_name.split(lsp_substring)
        extra_info = f"-{productType.lower()}-{productGroupId.lower()}"
        return split_parts[0] + lsp_substring + extra_info + split_parts[1]

    # Edge case 2: some swi collections are split into multiple manifests.
    swi_substring = "clms_global_swi"
    if product_index in SWI_SPLIT_MANIFEST:
        split_parts = manifest_name.split(swi_substring)
        extra_info = f"-{productGroupId.lower()}"
        return split_parts[0] + swi_substring + extra_info + split_parts[1]

    # Edge case 3: lcc collection is split into multiple manifests.
    if product_index in LCC_SPLIT_MANIFEST:
        return product_index.replace('yearly', str(year))


def write_results(results: List[str], filename: Union[str, Path]) -> None:
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    with open(filename, 'w+') as f:
        for href in results:
            f.write("%s\n" % href)


def are_equal_hrefs(other_href, product_href):
    parent_product, filename_product = os.path.split(product_href)
    parent_other, filename_other = os.path.split(other_href)

    pattern = r'(\d+_[A-Z]+_V\d+\.\d+\.\d+)'
    id_product = re.search(pattern, filename_product).group(1)
    id_other = re.search(pattern, filename_other).group(1)

    return parent_product == parent_other and id_product == id_other

def add_to_manifest(manifest_root_path: str, index: str, product_id: str) -> None:
    """
    Function that adds a product to the manifest file of a given index.

    :param manifest_root_path: the root path of the manifest files
    :param index: the index of the product
    :param product_id: the id of the product
    """
    # Only add to manifest if the index has a manifest.
    _do_non_bulk(manifest_root_path, index, product_id, add_to_products)

def bulk_add_to_manifests(manifest_root_path: str, input_path: str) -> None:
    """
    Function that adds all products from an input file located at input_path.

    :param manifest_root_path: the root path of the manifest files
    :param input_path: the path to the input file
    """

    _do_in_bulk(manifest_root_path, input_path, add_to_products)

def add_changing_hrefs_in_bulk(current_products, to_add):
    pattern = r'(\d+_[A-Z]+_V\d+\.\d+\.\d+)'
    to_add_ids = {re.search(pattern, os.path.split(product_href)[1]).group(1) : product_href for product_href in to_add}

    resulting_products = []
    for href in current_products:
        id_href = re.search(pattern, os.path.split(href)[1]).group(1)
        if id_href not in to_add_ids:
            resulting_products.append(href)
        else:
            resulting_products.append(to_add_ids[id_href])
            del to_add_ids[id_href]

    resulting_products.extend(to_add_ids.values())
    return resulting_products


def add_to_products(index, current_products, to_add):
    if index not in CHANGING_HREFS:
        current_products_set = set(current_products)
        current_products.extend([new_href for new_href in to_add if new_href not in current_products_set])
    else:
        current_products = add_changing_hrefs_in_bulk(current_products, to_add)
    return current_products

def remove_from_manifest(manifest_root_path: str, index: str, product_id: str) -> None:
    """
    Function that removes a product to the manifest file of a given index.

    :param manifest_root_path: the root path of the manifest files
    :param index: the index of the product
    :param product_id: the id of the product
    """
    # Only remove from manifest if the index has a manifest.
    _do_non_bulk(manifest_root_path, index, product_id, remove_from_products)

def bulk_remove_from_manifests(manifest_root_path: str, input_path) -> None:
    """
    Function that removes all products from an input file located at input_path.

    :param manifest_root_path: the root path of the manifest files
    :param input_path: the path to the input file
    """
    _do_in_bulk(manifest_root_path, input_path, remove_from_products)


def remove_from_products(index, current_products, to_remove):
    if index not in CHANGING_HREFS:
        resulting_products =  [product_href for product_href in current_products
                              if product_href not in to_remove]
    else:
        resulting_products =  [product_href for product_href in current_products # Remove products where href might have changed.
                              if not any([are_equal_hrefs(other_href, product_href) for other_href in to_remove])]
    return resulting_products

def _do_non_bulk(manifest_root_path: str, index: str, product_id: str, func):
    """
    Run the function func as a non-bulk function, usable functions are add_to_products and remove_from_products
    """
    if should_make_manifest(index, None):
        manifest_path, current_products, product_href = _get_details(manifest_root_path, index, product_id)
        resulting_products = func(index, current_products, [product_href])
        write_results(resulting_products, manifest_path)


def _do_in_bulk(manifest_root_path: str, input_path, func) -> None:
    """
    Run the function func as a bulk function, usable functions are add_to_products and remove_from_products
    """

    # Get the products per index from the input file.
    with open(input_path) as f:
        input = json.load(f)
        product_hrefs_per_file = {}
        path_to_index = {}
        catalogue = Catalogue(CatalogueConfig.from_environment(CatalogueEnvironment.CGLS))
        for product in input:
            manifest_path, product_href, index = _get_product_information(product["properties"]["parentIdentifier"],
                                         product["id"],
                                         manifest_root_path,
                                         catalogue)
            if manifest_path in product_hrefs_per_file:
                product_hrefs_per_file[manifest_path].append(product_href)
            else:
                path_to_index[manifest_path] = index
                product_hrefs_per_file[manifest_path] = [product_href]

    # Add the products to the manifest files, per file.
    for manifest_path in product_hrefs_per_file:
        index = path_to_index[manifest_path]
        hrefs = product_hrefs_per_file[manifest_path]
        if should_make_manifest(index, None):
            current_products = _get_current_products(manifest_path)
            resulting_products = func(index, current_products, hrefs)
            write_results(resulting_products, manifest_path)

def _get_details(manifest_root_path: Union[Path, str], index: str, product_id: str) -> Tuple[Path, List[str], str]:
    """
    Function that retrieves the path to the manifest file, the current products in the manifest and the href of the
    product.
    """
    manifest_path, product_href, _ = _get_product_information(index, product_id, manifest_root_path)
    current_products = _get_current_products(manifest_path)
    return manifest_path, current_products, product_href

def _get_current_products(manifest_path):
    with open(manifest_path, 'r') as f:
        return f.read().splitlines()


def _get_product_information(index, product_id, manifest_root_path, catalogue=None):
    """
    Function that retrieves the path to the manifest file, the current products in the manifest and the href of the
    product.
    """
    product = _get_product(index, product_id, catalogue)
    manifest_path = _manifest_path_of_product(manifest_root_path, index, product)

    # if not os.path.exists(manifest_path):
    #     raise Exception(f"Manifest file {manifest_path} does not exist!")

    product_href = product.properties["links"]["data"][0]["href"]
    return manifest_path, product_href, index

def _get_product(index: str, product_id: str, catalogue=None) -> Product:
    catalogue = Catalogue(CatalogueConfig.from_environment(CatalogueEnvironment.CGLS)) if catalogue is None else catalogue
    queried_products = list(catalogue.get_products(index, uid=product_id))
    if len(queried_products) == 1:
        return queried_products[0]
    elif len(queried_products) == 0:
        raise Exception(f"No product with id {product_id} were found!")
    else:
        raise Exception(f"Multiple products ({len(queried_products)}) with id {product_id} were found!")


def _manifest_path_of_product(manifest_root_path: Union[Path, str], index: str, product: Product) -> Path:
    product_type, product_group = None, None
    if index in LSP_SPLIT_MANIFEST or index in SWI_SPLIT_MANIFEST:
        product_type = product.properties["productInformation"]["productType"]
        product_group = product.properties["productInformation"]["productGroupId"]

    return get_manifest_path(manifest_root_path, index, productGroupId=product_group, productType=product_type)
