import datetime
import os

from oscars_mgmnt_clms.file_object import ObjectInfo
from oscars_mgmnt_clms.parser.geotiff import CLMSGlobalParser_GeoTIFF
import re


class LandCover_Parser_GeoTIFF(CLMSGlobalParser_GeoTIFF):
    # naming convention from the user manual: <TILE>_<SENSOR>_LC100_<AREA>_<VERSION>_<YEAR>-<MODE>_<LAYER>_<CRS>.tif
    product_regex = re.compile(
        r"(?:(?P<tile>[A-Za-z\d]+)_)?"
        r"(?P<sensor>[A-Za-z]+)"
        r"_"
        r"LC(?P<resolution>\d+)"
        r"_"
        r"global"
        r"_"
        r"(?P<version>v\d\.\d\.\d)"
        r"_"
        r"(?P<year>\d{4})-(?P<mode>\w+)"
        r"_"
        r"(?P<layer>[\w\-]+)"
        r"_"
        r"(?P<crs>[\w\-]+)"
    )

    def get_nominal_date(self, file_name: str):
        match = self.product_regex.match(file_name)
        return datetime.datetime.strptime(match.group("year"), "%Y")

    def get_file_title(self, filename: str) -> str:
        return self.product_regex.match(os.path.basename(filename)).group("layer")

    def specific_treatment(self, product_file_info: ObjectInfo, values: dict):
        self.fix_products(product_file_info)

        fn = os.path.abspath(product_file_info.get_product_path())
        file_name = os.path.basename(fn).rpartition(".")[0]
        regex_match = self.product_regex.match(file_name)

        values["id"] = file_name
        values["producttype"] = regex_match.group("layer")
        values["productVersion"] = regex_match.group("version")
        values["tileid"] = regex_match.group("tile") or "GLOBE"
        values["resolution"] = int(regex_match.group("resolution"))
        values[
            "conformsto"
        ] = f"https://www.opengis.net/def/crs/EPSG/0/{regex_match.group('crs').split('-')[-1]}"

        self.extract_raster_info(fn, values)

        return values["id"]

    def get_product_title_from_raster(self, raster, values: dict) -> str:
        tags = raster.tags()
        description = raster.tags(1).get("long_name")
        if values["tileid"] != "GLOBE":
            description += f" - {values['tileid']}"
        return f"{tags['title']} - {description}" if description else tags["title"]
